--
-- FarCaster
-- Written by: Trajan
-- Various fixes for patch 5.0.4 by Sargonan
--

--
-- Variable declarations:
--
tblSpellList = {};

local iCurrentTalentTree = 0;
local sCurrentTargetType = "Player";
local tblAllSpells = {}; -- All useable spells in spellbook
local iSpellCount = 0; -- Number of usable spells

local iPartyMembers = 0;
local iRaidMembers = 0;

local sUnitFrames = "";
local sRaidUnitFrames = "";

FarCaster_UpdateInterval = 0.1; -- How often the OnUpdate code will run (in seconds)
FCTimeSinceLastUpdate = 0.1;

BINDING_HEADER_FARCASTER = "FarCaster";
BINDING_NAME_DRAWSPELLBUTTONS = "Redraw spell buttons";
-----------------------------------------------------

--
-- Event handling
--
frm = CreateFrame("FRAME",UIParent);
frm:RegisterEvent("GROUP_ROSTER_UPDATE");
frm:RegisterEvent("ADDON_LOADED");
frm:RegisterEvent("PLAYER_ENTERING_WORLD");
frm:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED");

function frm:OnEvent(event, arg1)
	if (event == "ADDON_LOADED") and (arg1 == "Farcaster") then
		iCurrentTalentTree = GetSpecialization();

		if (iCurrentTalentTree ~= nil) and (tblSpellList["Tree"..iCurrentTalentTree] ~= nil) then
			if (tblSpellList["Tree"..iCurrentTalentTree] == nil) then
				tblSpellList["Tree"..iCurrentTalentTree] = {};
			end
		end
		
		if (tblSpellList["Settings"] == nil) then
			tblSpellList["Settings"] = {};
		end
		
		frmOptions:SetBackdropBorderColor(0.8, 0.6, 0.5, 0.9);

		SLASH_FARCASTER1 = "/farcast";
		SlashCmdList["FARCASTER"] = function(msg) btnFarCasterMiniMap_ShowOptions(); end
		
	elseif (event == "ACTIVE_TALENT_GROUP_CHANGED") then
		iCurrentTalentTree = GetSpecialization();
		createSpellButtons();
	elseif (event == "PLAYER_ENTERING_WORLD") then
		iCurrentTalentTree = GetSpecialization();
		setUnitFrames();
		createSpellButtons();
	elseif (event == "GROUP_ROSTER_UPDATE") then
        if IsInRaid() then
            raidChanged();
        else
            partyChanged();
        end
	end -- if (event)
end

-- This event is called X times per second, and is used to check if there's enough
-- mana to cast spell, if spell is on cooldown and if unit is within range.
function frm:OnUpdate(elapsed)
	FCTimeSinceLastUpdate = FCTimeSinceLastUpdate + elapsed;
	
	if (FCTimeSinceLastUpdate > FarCaster_UpdateInterval) then
		if (sUnitFrames == "") then
			setUnitFrames();
		end

		if (sUnitFrames ~= "" and iCurrentTalentTree ~= nil and tblSpellList["Tree"..iCurrentTalentTree] ~= nil) then
			-- Player
			UpdateUnitSpellButtons("Player","player");
			
			-- Pet
			if (UnitGUID("pet") ~= nil) then
				UpdateUnitSpellButtons("Pet","pet");
			end
			
			-- Target
			if (UnitGUID("target") ~= nil) then
				UpdateUnitSpellButtons("Target","target");
			end

			-- Focus
			if (UnitGUID("focus") ~= nil) then
				UpdateUnitSpellButtons("Focus","focus");
			end

			-- Party
			if (iPartyMembers > 0) then
				for iIndex = 1,iPartyMembers,1 do
					UpdateUnitSpellButtons("Party","party"..iIndex);
				end
			end
			
			-- Party pets
			if (iPartyMembers > 0) then
				for iIndex = 1,iPartyMembers,1 do
					if (UnitGUID("partypet"..iIndex) ~= nil) then
						UpdateUnitSpellButtons("Pet","partypet"..iIndex);
					end
				end
			end
			
			-- Raid frames
			for iIndex=1, iRaidMembers do
				if (UnitGUID("raid"..iIndex) ~= nil) then
					UpdateUnitSpellButtons("Raid","raid"..iIndex)
				end
			end
		end -- if sUnitFrames
		
		FCTimeSinceLastUpdate = 0; -- Reset timer
	end
end

function UpdateUnitSpellButtons(sTargetType,sTarget)
	if (tblSpellList["Tree"..iCurrentTalentTree][sTargetType] == nil) then return; end
	
	local iTIndex = 1;
	
	if (_G["texSpell_"..iTIndex..sTarget] == nil) then
		return;
	end
	
	-- Loop through spells for this target type
	for k, v in pairs (tblSpellList["Tree"..iCurrentTalentTree][sTargetType]["Spells"]) do
		local start, duration, enable = GetSpellCooldown(v);
			
		if (_G["cldSpellCD_"..iTIndex..sTarget]:IsVisible() == nil) then
			CooldownFrame_SetTimer(_G["cldSpellCD_"..iTIndex..sTarget], start, duration, enable)
		end
		
		_G["texSpell_"..iTIndex..sTarget]:SetVertexColor(1.0, 1.0, 1.0, 1.0);
		
		-- Check mana
		usable, nomana = IsUsableSpell(v);
		if (nomana == 1) then
			_G["texSpell_"..iTIndex..sTarget]:SetVertexColor(0.0, 0.0, 1.0);
		end
		
		-- Check range
		if (sTarget ~= "player") then
			UnitRangeCheck(v, sTarget, "texSpell_"..iTIndex..sTarget);
		end
		
		-- Check if spell has already been applied to target
		spellName, rank = GetSpellInfo(v);
		name, rank, icon, count, debuffType, duration, expirationTime, isMine, isStealable = UnitAura(sTarget, spellName);
		if (name ~= nil) then
			_G["texSpell_"..iTIndex..sTarget]:SetVertexColor(0.6, 0.6, 0.6, 0.4);
		end
		
		-- Check for debuffs
		for i=1,40 do
			name, rank, icon, count, debuffType = UnitDebuff(sTarget, i, 1);
			spellName, rank = GetSpellInfo(v);
			if (debuffType == "Magic") then
				if (v == 527 or v == 4987) then
					_G["texSpell_"..iTIndex..sTarget]:SetVertexColor(0.0, 1.0, 1.0);
				end
			elseif (debuffType == "Disease") then
				if (v == 528 or v == 4987) then
					_G["texSpell_"..iTIndex..sTarget]:SetVertexColor(0.9, 0.8, 0.2);
				end
			elseif (debuffType == "Poison") then
				if (v == 526 or v == 536 or v == 4987 or v == 2782) then
					_G["texSpell_"..iTIndex..sTarget]:SetVertexColor(0.0, 1.0, 0.0);
				end
			elseif (debuffType == "Curse") then
				if (v == 526 or v == 486 or v == 475 or v == 2782 or v == 51886) then
					_G["texSpell_"..iTIndex..sTarget]:SetVertexColor(1.0, 0.0, 1.0);
				end
			end
		end

		iTIndex = iTIndex + 1;
	end
end

-- Check if unit is in range of spell
function UnitRangeCheck(iSpellID, sUnit, sTexture)
	if (_G[sTexture] == nil) then return; end
	
	spellName, rank = GetSpellInfo(iSpellID);

	if (IsSpellInRange(spellName,sUnit) == 1) then
		_G[sTexture]:SetVertexColor(1.0, 1.0, 1.0);
		
		-- Check if player has enough mana to cast spell
		usable, nomana = IsUsableSpell(spellName);
		if (nomana == 1) then
			_G[sTexture]:SetVertexColor(0.0, 0.0, 1.0);
		end
	else
		_G[sTexture]:SetVertexColor(1.0, 0.0, 0.0);
	end
end

frm:SetScript("OnEvent", frm.OnEvent);
frm:SetScript("OnUpdate", frm.OnUpdate);

function partyChanged()
	iPartyMembers = GetNumSubgroupMembers();

	createSpellButtons();
end

function raidChanged()
	iRaidMembers = GetNumGroupMembers();
		
	if (sRaidUnitFrames == "") then
		setUnitFrames();
	end
	
	createSpellButtons();
end

--

-- Automatically detect which unit frame system is used
function setUnitFrames()
	for k, v in pairs (tblFarCaster_LockToFramesAvail) do
		if (_G[tblFarCaster_LockToFrames[v]["FarCasterPlayerBar"]] ~= nil) then
			sUnitFrames = v;
		end
	end

	for k, v in pairs (tblFarCaster_RaidLockToFramesAvail) do
		local fr = _G[tblFarCaster_RaidLockToFrames[v]["RaidGroup"].."1"..tblFarCaster_RaidLockToFrames[v]["RaidGroupMembers"].."1"];
		if (_G[tblFarCaster_RaidLockToFrames[v]["RaidGroupMembers"].."1"] ~= nil and _G[tblFarCaster_RaidLockToFrames[v]["RaidGroupMembers"].."1"]:IsVisible()) or (fr ~= nil and fr:IsVisible()) then
			if (sRaidUnitFrames == "") then
				sRaidUnitFrames = v;
			end
		end
	end
end

-- Create spell list tables if they do not already exist
function initialiseSpellListTable()
	if (tblSpellList["Tree"..iCurrentTalentTree] == nil) then
		tblSpellList["Tree"..iCurrentTalentTree] = {};
	end
	
	if (tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType] == nil) then
		tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType] = {};
	end
end

-- Parse player's spellbook and build a list of available spells
function getSpells()
    -- Only retrieve available spells once
    if (iSpellCount == 0) then
      local i = GetNumSpellTabs();
      local n, iSpellId, iIndex, cSpellId, cIndex, cadded;
      while true do
           local name, texture, offset, numSpells = GetSpellTabInfo(i);
          if not name then
              break
          end
          n = numSpells + offset
          i = i + 1
       end
      for iIndex = 1, n do
       skillType, iSpellId = GetSpellBookItemInfo(iIndex, "player")
       if (skillType == "SPELL" and IsPlayerSpell(iSpellId) ) or skillType == "FLYOUT" then
           skillType, iSpellId = GetSpellBookItemInfo(iIndex, 'spell');
               if (iSpellId ~= nil) then
                       isPassive = IsPassiveSpell(iIndex, BOOKTYPE_SPELL);
                       spellName, rank = GetSpellInfo(iSpellId);

                       if (spellName ~= nil) then
                             local sTexture = GetSpellBookItemTexture(spellName);

                   if (isPassive == nil) and (sTexture ~= nil) then
                      -- added loop check to see if spell already added, I'm sure there
                      -- must be a better way to do this. Sargonan
                      cadded = nil;
                                    for cIndex = 1,iSpellCount,1 do
                         cSpellId = tblAllSpells[cIndex];
                         if cSpellId == iSpellId then
                            cadded = 1;
                         end
                                    end                                    
                                    if (cadded == nil) then
                                      iSpellCount = iSpellCount + 1;
                                    table.insert(tblAllSpells,iSpellCount,iSpellId);
                                 end
                            end
                        end   
                   end
             end  
             iIndex = iIndex + 1;
        end -- for
    end
end

-- Create a vertical row of all available spells in the options panel
function drawSpellList()
	-- Loop through all available spells
	for iIndex = 1,iSpellCount,1 do
		if (_G["btnSpell"..iIndex] == nil) then
			local iSpellId = tblAllSpells[iIndex];
			spellName, rank = GetSpellInfo(iSpellId);
			local sTexture = GetSpellBookItemTexture(spellName);

			local btnSpell = CreateFrame("Button","btnSpell"..iIndex,frmSpellList);

			-- Get spell icon
			local texture = btnSpell:CreateTexture(nil, "BACKGROUND");
			texture:SetTexture(sTexture);
			texture:SetAllPoints("btnSpell"..iIndex);

			btnSpell:SetPoint("TOPLEFT", 20, 0 - (iIndex*25));
			btnSpell:SetWidth(24);
			btnSpell:SetHeight(24);

			-- Add tooltip to icon
			local iSID = iSpellId;
			btnSpell:SetScript("OnEnter",function()
				GameTooltip:SetOwner(btnSpell, "ANCHOR_BOTTOMRIGHT");
				GameTooltip:SetHyperlink(GetSpellLink(iSID)) 
				GameTooltip:Show();
			end)
			btnSpell:SetScript("OnLeave",function()
				GameTooltip:Hide();
			end)

			btnSpell:Show();
			
			-- Places a checkbox to the right of the spell icon
			local chkSpell = CreateFrame("CheckButton", "chkSpell"..iIndex, frmSpellList, "OptionsCheckButtonTemplate");
			chkSpell:SetWidth(20);
			chkSpell:SetHeight(20);
			chkSpell:SetPoint("TOPLEFT", 50, 0 - (iIndex*25));
			chkSpell:SetNormalTexture("Interface\\Buttons\\UI-CheckBox-Up");
			chkSpell:SetPushedTexture("Interface\\Buttons\\UI-CheckBox-Down");
			chkSpell:SetHighlightTexture("Interface\\Buttons\\UI-CheckBox-Highlight", "ADD");
			chkSpell:SetCheckedTexture("Interface\\Buttons\\UI-CheckBox-Check");

			local iChk = iIndex;
			chkSpell:SetScript("OnClick",function()
				chkSpell_Click(iChk);
			end)
			
			chkSpell:Show()
		end
	end
end

-- Mark all spells that have been selected for the current target type
function updateSpellCheckboxes()
	for iIndex = 1,iSpellCount,1 do
		local chkSpell = _G["chkSpell"..iIndex];
		chkSpell:SetChecked(false);

		if (iCurrentTalentTree == nil) then
			iCurrentTalentTree = GetSpecialization();
		end

		initialiseSpellListTable();
		if (tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"] == nil) then
			tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"] = {};
		end

		if(tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"] ~= nil) then
			for k, v in pairs (tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"]) do
				if (v == tblAllSpells[iIndex]) then
					chkSpell:SetChecked(true);
				end
			end
		end --end if
	end -- end for
end

-- Handles clickevent on the checkboxes in the list of available spells
function chkSpell_Click(chk)
	local chkSpell = _G["chkSpell"..chk]; -- Get a reference to the checkbox that was clicked
	
	if (tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType] == nil) then
		tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType] = { };
	end
	
	if (chkSpell:GetChecked() == 1) then
		-- Make sure table exists
		if (tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"] ~= nil) then
			-- First check if spell already exists in table
			local bFound = false;
			for k, v in pairs (tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"]) do
				if (v == tblAllSpells[chk]) then
					bFound = true;
				end
			end
			if (bFound == false) then
				table.insert(tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"],tblAllSpells[chk]);
			end
		else
			-- Table does not exist yet, so it must be initialised before we can insert a new spell.
			tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"] = {};
			table.insert(tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"],tblAllSpells[chk]);
		end
	else
		-- Delete spell
		if (tblSpellList ~= nil) then
			local iIndex = 1;
			-- Loop through spell list to find index of spell to delete
			while tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"][iIndex] ~= nil do
				if (tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"][iIndex] == tblAllSpells[chk]) then
					table.remove(tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"],iIndex);
				end
				iIndex = iIndex + 1;
			end
		end -- end if
	end
	
	drawSelectedSpells();
	createSpellButtons();
end

-- Draws a vertical row of spell icons based on
-- current talent spec and target type. The user
-- can change the order of the spells by clicking
-- arrow buttons next to each spell.
function drawSelectedSpells()
	-- Hide all spell icons
	local iIndex = 1;
	while (_G["btnSelectedSpell"..iIndex] ~= nil) do
		btnTemp = _G["btnSelectedSpell"..iIndex];
		btnTemp:Hide();
		--if(iIndex > 1) then
			btnTemp = _G["btnUp"..iIndex];
			btnTemp:Hide();
		--end
		btnTemp = _G["btnDown"..iIndex];
		btnTemp:Hide();
		iIndex = iIndex + 1;
	end
	
	initialiseSpellListTable();
	
	if(tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"] ~= nil) then
		iIndex = 1;
		for k, v in pairs (tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"]) do
			spellName, rank = GetSpellInfo(v);
			local sTexture = GetSpellBookItemTexture(spellName);

			local btnSpell;
			if(_G["btnSelectedSpell"..iIndex] == nil) then
				btnSpell = CreateFrame("Button","btnSelectedSpell"..iIndex,frmSelectedSpells);
			else
				btnSpell = _G["btnSelectedSpell"..iIndex];
			end

			-- Get spell icon
			local texture;
			if(_G["texSelectedSpell"..iIndex] == nil) then
				texture = btnSpell:CreateTexture("texSelectedSpell"..iIndex, "BACKGROUND");
			else
				texture = _G["texSelectedSpell"..iIndex];
			end
			texture:SetTexture(sTexture);
			texture:SetAllPoints("btnSelectedSpell"..iIndex);

			btnSpell:SetPoint("TOPLEFT", 50, 0 - (iIndex*28));
			btnSpell:SetWidth(26);
			btnSpell:SetHeight(26);

			-- Add tooltip to icon
			local iSID = v;
			btnSpell:SetScript("OnEnter",function()
				GameTooltip:SetOwner(btnSpell, "ANCHOR_BOTTOMRIGHT");
				GameTooltip:SetHyperlink(GetSpellLink(iSID)) 
				GameTooltip:Show();
			end)
			btnSpell:SetScript("OnLeave",function()
				GameTooltip:Hide();
			end)

			btnSpell:Show();
			
			-- The arrow buttons move spells up and down.
			-- Up arrow
			if (_G["btnUp"..iIndex] == nil) then
				local btnUp = CreateFrame("Button", "btnUp"..iIndex, frmSelectedSpells, "UIPanelScrollUpButtonTemplate");
				btnUp:SetWidth(16)
				btnUp:SetHeight(16)
				btnUp:SetPoint("TOPLEFT", 90, -3 - (iIndex*28))
				local i = iIndex;
				btnUp:SetScript("OnClick",function()
					btnUp_Click(i);
				end)
				btnUp:Show();
			else
				local btnUp = _G["btnUp"..iIndex];
				btnUp:Show();
			end
			-- Down arrow
			if (_G["btnDown"..iIndex] == nil) then
				local btnDown = CreateFrame("Button", "btnDown"..iIndex, frmSelectedSpells, "UIPanelScrollDownButtonTemplate");
				btnDown:SetWidth(16)
				btnDown:SetHeight(16)
				btnDown:SetPoint("TOPLEFT", 115, -3 - (iIndex*28))
				local i = iIndex;
				btnDown:SetScript("OnClick",function()
					btnDown_Click(i);
				end)
				btnDown:Show();
			else
				local btnDown = _G["btnDown"..iIndex];
				btnDown:Show();
			end
			
			iIndex = iIndex + 1;
		end -- for each spell
	end -- if
end

-- Move spell up
function btnUp_Click(iIndex)
	local iTemp = tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"][iIndex-1];
	tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"][iIndex-1] = tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"][iIndex];
	tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"][iIndex] = iTemp;
	
	drawSelectedSpells();
	createSpellButtons();
end

-- Move spell down
function btnDown_Click(iIndex)
	local iTemp = tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"][iIndex+1];
	tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"][iIndex+1] = tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"][iIndex];
	tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Spells"][iIndex] = iTemp;
	
	drawSelectedSpells();
	createSpellButtons();
end

-- Creates a slider that changes the size of the spell buttons
function drawIconScaleSlider()
	local sdrIconScale;
	if (_G["sdrIconScale"] == nil) then
		sdrIconScale = CreateFrame("Slider", "sdrIconScale", frmScaleAndPosition, "OptionsSliderTemplate");
	else
		sdrIconScale = _G["sdrIconScale"];
	end

	sdrIconScale:SetMinMaxValues(8, 64)
	
	initialiseSpellListTable();
	if (tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["IconSize"] == nil) then
		tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["IconSize"] = 24;
	end
	
	if (tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["IconSize"] > 0) then
		sdrIconScale:SetValue(tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["IconSize"]);
	else
		sdrIconScale:SetValue(32);
	end
	sdrIconScale:SetWidth(200)
	sdrIconScale:SetHeight(20)
	sdrIconScale:SetOrientation('HORIZONTAL')
	sdrIconScale:SetPoint("TOPLEFT", frmScaleAndPosition, "TOPLEFT", 50, -50)
	_G[sdrIconScale:GetName() .. 'Low']:SetText('Small');
	_G[sdrIconScale:GetName() .. 'High']:SetText('Large');
	_G[sdrIconScale:GetName() .. 'Text']:SetText('Icon size');
	
	sdrIconScale:SetScript("OnValueChanged",function()
		sdrIconScale_Changed();
	end)
	
	sdrIconScale:Show();
end

-- Button size slider change event
function sdrIconScale_Changed()
	local sdrIconScale = _G["sdrIconScale"];
	local iValue = sdrIconScale:GetValue();
	iValue = math.floor(iValue + 0.5);
	
	tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["IconSize"] = iValue;
	
	createSpellButtons();
end

-- Create input boxes for X and Y position values
function drawPositionEditBoxes()
	local txtXPos;
	if (_G["txtXPos"] == nil) then
		txtXPos = CreateFrame("EditBox", "txtXPos", frmScaleAndPosition, "InputBoxTemplate");
		
		txtXPos:SetAutoFocus(false);
		
		txtXPos:SetWidth(40); -- to see three digits instead of two
		txtXPos:SetHeight(20);
		txtXPos:SetPoint("TOPLEFT", frmScaleAndPosition, "TOPLEFT", 300, -50);

		txtXPos:SetScript("OnTextChanged",function()
			txtXPos_Changed();
		end)

		-- Text label
		local lblXPos = frmScaleAndPosition:CreateFontString("lblXPos", frmScaleAndPosition, "GameFontWhite");
		lblXPos:SetPoint("TOPLEFT", _G["txtXPos"], "TOPLEFT", -20, 0);
		lblXPos:SetText("X:");
	else
		txtXPos = _G["txtXPos"];
	end
	txtXPos:Show();
	lblXPos:Show();
	
    if (iCurrentTalentTree ~= nil) then
	    if(tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["X"] ~= nil) then
		    txtXPos:SetText(tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["X"]);
	    end
    end
	
	--------------------------------------------------------------------
	
	local txtYPos;
	if (_G["txtYPos"] == nil) then
		txtYPos = CreateFrame("EditBox", "txtYPos", frmScaleAndPosition, "InputBoxTemplate");
		
		txtYPos:SetAutoFocus(false);
		
		txtXPos:SetWidth(40); -- to see three digits instead of two
		txtYPos:SetHeight(20);
		txtYPos:SetPoint("TOPLEFT", frmScaleAndPosition, "TOPLEFT", 300, -70);

		txtYPos:SetScript("OnTextChanged",function()
			txtYPos_Changed();
		end)
		
		-- Text label
		local lblYPos = frmScaleAndPosition:CreateFontString("lblYPos", frmScaleAndPosition, "GameFontWhite");
		lblYPos:SetPoint("TOPLEFT", _G["txtYPos"], "TOPLEFT", -20, 0);
		lblYPos:SetText("Y:");
	else
		txtYPos = _G["txtYPos"];
	end
	txtYPos:Show();
	lblYPos:Show();
	
    if (iCurrentTalentTree ~= nil) then
	    if(tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Y"] ~= nil) then
		    txtYPos:SetText(tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Y"]);
	    end
    end
end

function txtXPos_Changed()
	tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["X"] = _G["txtXPos"]:GetNumber();
	createSpellButtons();
end

function txtYPos_Changed()
	tblSpellList["Tree"..iCurrentTalentTree][sCurrentTargetType]["Y"] = _G["txtYPos"]:GetNumber();
	createSpellButtons();
end

-- User clicked a tab in the option window
function btnChangeTarget_Click(sTarget)
	frmSpellSettings:Show();
	frmGeneralSettings:Hide();
	sCurrentTargetType = sTarget;
	fntHeading:SetText(sTarget.." spells"); 
	updateSpellCheckboxes();
	drawSelectedSpells();
	drawIconScaleSlider();
	drawPositionEditBoxes();
end

-- General settings are handled here
function btnSettings_Click()
	frmSpellSettings:Hide();
	frmGeneralSettings:Show();
	fntHeading:SetText("General settings");
	
	-- Create a checkbox for turning tooltips on/off
	local chkTooltip;
	if (_G["chkTooltip"] == nil) then
		chkTooltip = CreateFrame("CheckButton", "chkTooltip", frmGeneralSettings, "OptionsCheckButtonTemplate");
		chkTooltip:SetWidth(20);
		chkTooltip:SetHeight(20);
		chkTooltip:SetPoint("TOPLEFT", 25, -45);
		chkTooltip:SetNormalTexture("Interface\\Buttons\\UI-CheckBox-Up");
		chkTooltip:SetPushedTexture("Interface\\Buttons\\UI-CheckBox-Down");
		chkTooltip:SetHighlightTexture("Interface\\Buttons\\UI-CheckBox-Highlight", "ADD");
		chkTooltip:SetCheckedTexture("Interface\\Buttons\\UI-CheckBox-Check");

		local iChk = iIndex;
		chkTooltip:SetScript("OnClick",function()
			chkTooltip_Click();
		end)
	else
		chkTooltip = _G["chkTooltip"];
	end
	
	-- Tooltip checkbox text
	local fntTooltip = frmGeneralSettings:CreateFontString(nil, "OVERLAY", "GameFontNormal");
	fntTooltip:SetPoint("TOPLEFT", _G["chkTooltip"], "TOPLEFT", 30, -4);
	fntTooltip:SetText("Show tooltips on spell bars");
	fntTooltip:Show() 
	
	if (tblSpellList["Settings"]["ShowTooltip"] == "true") then
		chkTooltip:SetChecked(true);
	end
	chkTooltip:Show()
end

function chkTooltip_Click()
	if (tblSpellList["Settings"] == nil) then
		tblSpellList["Settings"] = { };
	end
	if (_G["chkTooltip"]:GetChecked() == 1) then
		tblSpellList["Settings"]["ShowTooltip"] = "true";
	else
		tblSpellList["Settings"]["ShowTooltip"] = "false";
	end
	
	createSpellButtons();
end

function btnFarCasterCloseWindow_Click()
	frmOptions:Hide();
end


--
-- Unit frame related functions
--
-- These are the functions that actually create and position
-- the spell buttons on the unit frames.
--
--

function createSpellButtons()
	-- Make sure there's an active talent tree and that spells have been assigned to that tree
	if (iCurrentTalentTree == nil or tblSpellList["Tree"..iCurrentTalentTree] == nil) then
		return;
	end
	
	setUnitFrames();

	-- Check if the type of unit frames system has been set
	if (sUnitFrames ~= "") then
		-- Player frame
		local frmAnchor = _G[tblFarCaster_LockToFrames[sUnitFrames]["FarCasterPlayerBar"]];
		drawSpellButtonBar("Player",frmAnchor,"player");

		-- Player pet frame
		if (tblFarCaster_LockToFrames[sUnitFrames]["FarCasterPetBar"] ~= nil) then
			local frmAnchor = _G[tblFarCaster_LockToFrames[sUnitFrames]["FarCasterPetBar"]];
			drawSpellButtonBar("Pet",frmAnchor,"pet");
		end
		
		-- Target frame
		if (tblFarCaster_LockToFrames[sUnitFrames]["FarCasterTargetBar"] ~= nil) then
			local frmAnchor = _G[tblFarCaster_LockToFrames[sUnitFrames]["FarCasterTargetBar"]];
			drawSpellButtonBar("Target",frmAnchor,"target");
		end
		
		-- Focus frame
		if (tblFarCaster_LockToFrames[sUnitFrames]["FarCasterFocusBar"] ~= nil) then
			local frmAnchor = _G[tblFarCaster_LockToFrames[sUnitFrames]["FarCasterFocusBar"]];
			drawSpellButtonBar("Focus",frmAnchor,"focus");
		end

		-- Party frames
		if (iPartyMembers > 0) then
			for iIndex = 1,iPartyMembers,1 do
				local frmAnchor = _G[tblFarCaster_LockToFrames[sUnitFrames]["FarCasterPartyBar"]..iIndex];
				local sTarget = "party"..iIndex;
				drawSpellButtonBar("Party",frmAnchor,sTarget);
			end
		end
		
		-- Party pet frames
		if (iPartyMembers > 0) then
			if (tblFarCaster_LockToFrames[sUnitFrames]["FarCasterPartyPetBar1"] ~= nil) then
				if (_G[tblFarCaster_LockToFrames[sUnitFrames]["FarCasterPartyPetBar1"]] ~= nil) then
					drawSpellButtonBar("Pet",_G[tblFarCaster_LockToFrames[sUnitFrames]["FarCasterPartyPetBar1"]],"partypet1");
				end
				if (_G[tblFarCaster_LockToFrames[sUnitFrames]["FarCasterPartyPetBar2"]] ~= nil) then
					drawSpellButtonBar("Pet",_G[tblFarCaster_LockToFrames[sUnitFrames]["FarCasterPartyPetBar2"]],"partypet2");
				end
				if (_G[tblFarCaster_LockToFrames[sUnitFrames]["FarCasterPartyPetBar3"]] ~= nil) then
					drawSpellButtonBar("Pet",_G[tblFarCaster_LockToFrames[sUnitFrames]["FarCasterPartyPetBar3"]],"partypet3");
				end
				if (_G[tblFarCaster_LockToFrames[sUnitFrames]["FarCasterPartyPetBar4"]] ~= nil) then
					drawSpellButtonBar("Pet",_G[tblFarCaster_LockToFrames[sUnitFrames]["FarCasterPartyPetBar4"]],"partypet4");
				end
			end
		end
	end

	if sRaidUnitFrames == "" then
		setUnitFrames();
	end

	-- Raid frames
	if (sRaidUnitFrames ~= "") then -- No raid groups
		local iRaidMaxGroups = tblFarCaster_RaidLockToFrames[sRaidUnitFrames]["MaxRaidGroups"];
		local iRaidMaxMembers = tblFarCaster_RaidLockToFrames[sRaidUnitFrames]["MaxRaidGroupMembers"];

		if (tblFarCaster_RaidLockToFrames[sRaidUnitFrames]["RaidGroup"] == "") then
			for i=1, iRaidMaxMembers do
				local frmRaidMember = _G[tblFarCaster_RaidLockToFrames[sRaidUnitFrames]["RaidGroupMembers"]..i];

				if(frmRaidMember and frmRaidMember:IsVisible()) then
					if(frmRaidMember.unit ~= nil or frmRaidMember.partyid ~= nil) then
						local frmAnchor = frmRaidMember;
						local sTarget = frmRaidMember.unit;
						drawSpellButtonBar("Raid",frmAnchor,sTarget);
					end
				end
			end
		else
			for i = 1, iRaidMaxGroups do -- iterate through raid groups
				local frmRaidGroup = _G[tblFarCaster_RaidLockToFrames[sRaidUnitFrames]["RaidGroup"]..i];
				if (frmRaidGroup and frmRaidGroup:IsShown()) then
					for ii=1, iRaidMaxMembers do
						local frmRaidMember = _G[frmRaidGroup:GetName()..tblFarCaster_RaidLockToFrames[sRaidUnitFrames]["RaidGroupMembers"]..ii];

						if(frmRaidMember) then
							if(frmRaidMember.unit ~= nil or frmRaidMember.partyid ~= nil) then
								local frmAnchor = frmRaidMember;
								local sTarget = frmRaidMember.unit;
								if (sTarget == nil) then
									sTarget = frmRaidMember.partyid;
								end
								drawSpellButtonBar("Raid",frmAnchor,sTarget);
							end
						end
					end
				end -- if
			end -- for
		end -- if raid group
	end -- if
end

-- Draws a row of spell icons based on input parameters.
function drawSpellButtonBar(sTargetType,frmAnchor,sTarget)
	if (tblSpellList["Tree"..iCurrentTalentTree][sTargetType] == nil) then
		return;
	end
	if (tblSpellList["Tree"..iCurrentTalentTree][sTargetType]["Spells"] == nil) then
		return;
	end

	if (frmAnchor == nil) then return; end

	if (UnitAffectingCombat(sTarget) ~= 1) then	
		-- Hide all spell icons first
		local iIndex = 1;
		while (_G["btnSpell_"..iIndex..sTarget] ~= nil) do
			btnTemp = _G["btnSpell_"..iIndex..sTarget];
			btnTemp:Hide();
			iIndex = iIndex + 1;
		end

		if (tblSpellList["Tree"..iCurrentTalentTree][sTargetType]["Spells"] ~= nil) then
			iIndex = 1;
			-- Loop through spells and create buttons
			for k, v in pairs (tblSpellList["Tree"..iCurrentTalentTree][sTargetType]["Spells"]) do
				local btnSpell;
				if(_G["btnSpell_"..iIndex..sTarget] == nil) then
					if (UnitAffectingCombat(sTarget) == 1) then break; end
					btnSpell = CreateFrame("Button","btnSpell_"..iIndex..sTarget,frmAnchor,"SecureActionButtonTemplate");
				else
					btnSpell = _G["btnSpell_"..iIndex..sTarget];
				end
				
				-- Size and position
				local iIconSize = tblSpellList["Tree"..iCurrentTalentTree][sTargetType]["IconSize"];
				if (iIconSize == nil) then
					iIconSize = 32;
				end
				
				if (tblSpellList["Tree"..iCurrentTalentTree][sTargetType]["X"] == nil) then
					tblSpellList["Tree"..iCurrentTalentTree][sTargetType]["X"] = 200;
				end
				if (tblSpellList["Tree"..iCurrentTalentTree][sTargetType]["Y"] == nil) then
					tblSpellList["Tree"..iCurrentTalentTree][sTargetType]["Y"] = -20;
				end
				
				btnSpell:SetPoint("TOPLEFT", tblSpellList["Tree"..iCurrentTalentTree][sTargetType]["X"] + (iIndex*(iIconSize+2)), tblSpellList["Tree"..iCurrentTalentTree][sTargetType]["Y"]);
				btnSpell:SetWidth(iIconSize);
				btnSpell:SetHeight(iIconSize);
				
				btnSpell:SetAttribute("type", "spell" );
				btnSpell:SetAttribute("spell", v );
				btnSpell:SetAttribute("unit", sTarget );

				btnSpell:SetPushedTexture("Interface\\Buttons\\UI-Quickslot-Depress");

				-- Get and set spell icon
				spellName, rank = GetSpellInfo(v);
				local sTexture = GetSpellBookItemTexture(spellName);
				local texture;
				if(_G["texSpell_"..iIndex..sTarget] == nil) then
					texture = btnSpell:CreateTexture("texSpell_"..iIndex..sTarget, "BACKGROUND");
				else
					texture = _G["texSpell_"..iIndex..sTarget];
				end
				texture:SetTexture(sTexture);
				texture:SetAllPoints("btnSpell_"..iIndex..sTarget);

				btnSpell:Show();

				-- Cooldown frame
				if(_G["cldSpellCD_"..iIndex..sTarget] == nil) then
					cldSpellCD = CreateFrame("Cooldown","cldSpellCD_"..iIndex..sTarget,btnSpell,"CooldownFrameTemplate");
				else
					cldSpellCD = _G["cldSpellCD_"..iIndex..sTarget];
				end
				btnSpell.cooldown = cldSpellCD;
				cldSpellCD:SetWidth(iIconSize);
				cldSpellCD:SetHeight(iIconSize);
				cldSpellCD:SetPoint("CENTER", 0, -1);

				cldSpellCD:Hide();
				
				-- Add tooltip to button if set to "true" in settings
				if (tblSpellList["Settings"]["ShowTooltip"] == "true") then
					local iSID = v;
					btnSpell:SetScript("OnEnter",function()
						GameTooltip:SetOwner(btnSpell, "ANCHOR_BOTTOMRIGHT");
						GameTooltip:SetHyperlink(GetSpellLink(iSID)) 
						GameTooltip:Show();
					end)
					btnSpell:SetScript("OnLeave",function()
						GameTooltip:Hide();
					end)
				else
					btnSpell:SetScript("OnEnter", nil);
				end
				
				iIndex = iIndex + 1;
			end -- for
		end -- if
	end -- if combat
end


function frmOptions_OnLoad()
	--DEFAULT_CHAT_FRAME:AddMessage("FarCaster: Loaded!")
end

--
-- Mini map icon functions
--

function btnFarCasterMiniMap_OnLoad()
	btnFarCasterMiniMap:SetScript("OnDragStart", btnFarCasterMiniMap.StartMoving)
	btnFarCasterMiniMap:SetScript("OnDragStop", btnFarCasterMiniMap.StopMovingOrSizing)
	btnFarCasterMiniMap:RegisterForDrag("LeftButton","RightButton")
end

function btnFarCasterMiniMap_OnEvent()
end

function btnFarCasterMiniMap_ShowOptions()
	if (frmOptions) then
		if(frmOptions:IsVisible()) then
			frmOptions:Hide();
		else
			if (iCurrentTalentTree == nil) then
				iCurrentTalentTree = GetSpecialization();
			end
			
			-- Options window will not open before an active talent tree has been found
			if (iCurrentTalentTree ~= nil) then
				frmOptions:Show();

				getSpells();
				drawSpellList();
				updateSpellCheckboxes();
				drawSelectedSpells();
				drawIconScaleSlider();
				drawPositionEditBoxes();
				fntHeading:SetText(sCurrentTargetType.." spells");
			end
		end
	end
end